
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <zlib.h>
#include <string.h>

const double Ya = 1, Qa = 5, Ia = 5, phi = 0;


#define coefs lowpass
#include "lowpass-3MHz.c"
#undef coefs

enum { lp_size = sizeof(lowpass) / sizeof(lowpass[0]) };

#define coefs bandpass
#include "bandpass-3MHz-4.2MHz.c"
#undef coefs

enum { bp_size = sizeof(bandpass) / sizeof(bandpass[0]) };

#define coefs lowpass_1
#include "lowpass-1MHz.c"
#undef coefs

enum { lp1_size = sizeof(lowpass_1) / sizeof(lowpass_1[0]) };


typedef unsigned char u8;

u8 data[720*486];
u8 img[720*486*3];

enum { tx=720, ty=486 };

void w32(unsigned char *p, int l)
{
  p[0] = l>>24;
  p[1] = l>>16;
  p[2] = l>>8;
  p[3] = l;
}

void wchunk(int fd, unsigned int type, unsigned char *p, unsigned int l)
{
  unsigned char v[8];
  unsigned int crc;
  w32(v, l);
  w32(v+4, type);
  write(fd, v, 8);
  crc = crc32(0, v+4, 4);
  if(l) {
    write(fd, p, l);
    crc = crc32(crc, p, l);
  }
  w32(v, crc);
  write(fd, v, 4);
}

void save(const char *fname)
{
  int x,y,fd;
  unsigned char *image, *cdata;

  image = (unsigned char *)malloc((tx*3+1)*ty);
  cdata = (unsigned char *)malloc((int)((tx*3+1)*ty*1.1+12));

  fd=open(fname, O_WRONLY|O_CREAT|O_TRUNC, 0666);
  if(fd < 0) {
    perror("open write");
    exit(1);
  }
  write(fd, "\x89\x50\x4E\x47\x0D\x0A\x1A\x0A", 8);

  w32(cdata, tx);
  w32(cdata+4, ty);
  cdata[8] = 8;
  cdata[9] = 2;
  cdata[10] = 0;
  cdata[11] = 0;
  cdata[12] = 0;

  wchunk(fd, 0x49484452L, cdata, 13); // IHDR

  unsigned char *ddata = image;

  for(y=0;y<ty;y++) {
    *ddata++ = 0;
    for(x=0;x<tx;x++) {
      const u8 *col = img+3*x+3*720*y;

      *ddata++ = *col++;
      *ddata++ = *col++;
      *ddata++ = *col++;
    }
  }

  unsigned long sz = (int)((tx*3+1)*ty*1.1+12);
  compress(cdata, &sz, image, (tx*3+1)*ty);
  wchunk(fd, 0x49444154L, cdata, sz); // IDAT
  wchunk(fd, 0x49454E44L, 0, 0); // IEND
  close(fd);
}

void load(const char *fname)
{
  int fd = open(fname, O_RDONLY);
  if(fd < 0) {
    perror("open read");
    exit(1);
  }

  read(fd, data, tx*ty);
  close(fd);
}

void applyfilt(const double *src, double *dest, int sz, const double *f, int fsz)
{
  int x, i;
  int mp = fsz/2;
  for(x=0; x<sz; x++) {
    double r = 0;
    for(i=0; i<fsz; i++) {
      int k = x+mp-i;
      if(k>=0 && k < sz)
	r += src[k]*f[i];
    }
    dest[x] = r;
  }
}

u8 clamp(double v)
{
  if(v<0)
    return 0;
  if(v>255)
    return 255;
  return (u8)v;
}

void decode_line(const u8 *src, u8 *dest, double phase)
{
  const double csub = 2 * M_PI * 3579545.0 / 13500000.0;

  double S[tx];
  double Y[tx], C[tx], I[tx], Q[tx], tmp[tx];
  int x;

  for(x=0; x<tx; x++)
    S[x] = src[x];

  // 1- filtering
  applyfilt(S, Y, tx, lowpass, lp_size);
  applyfilt(S, C, tx, bandpass, bp_size);

  // 2- QAM demodulation
  for(x=0; x<tx; x++)
    tmp[x] = C[x]*cos(csub*x+phase);
  applyfilt(tmp, I, tx, lowpass_1, lp1_size);

  for(x=0; x<tx; x++)
    tmp[x] = C[x]*sin(csub*x+phase);
  applyfilt(tmp, Q, tx, lowpass_1, lp1_size);


  // YIQ -> RGB

  for(x=0; x<tx; x++) {
    double y = Ya*Y[x];
    double i = Ia*I[x];
    double q = Qa*Q[x];

    double r = y + 0.956*i + 0.621*q;
    double g = y - 0.272*i - 0.647*q;
    double b = y - 1.107*i + 1.704*q;

    *dest++ = clamp(r);
    *dest++ = clamp(g);
    *dest++ = clamp(b);
  }
}

void decode(void)
{
  int field;
  for(field=0; field<2; field++) {
    int line;
    int phi_inv = 0;
    u8 *dest = img + field*3*tx;
    for(line=0; line<ty/2; line++) {
      const u8 *src = data+720*(2*line+field);
      decode_line(src, dest, phi + (phi_inv ? M_PI : 0));
      phi_inv = !phi_inv;
      dest += 2*3*tx;
    }
  }
}

int main(int argc, char **argv)
{
  if(argc != 3) {
    fprintf(stderr, "Usage:\n%s file.bin file.png\n", argv[0]);
    exit(1);
  }

  load(argv[1]);
  decode();
  save(argv[2]);
  return 0;
}
